@@ -34,53 +34,25 @@
 #include "tclap/CmdLine.h"
 #include "tclap/Arg.h"
 
-//#define NUM_IBDS 101
-//#define NUM_MIX_FRACS 51
-#define MAX_Q 100
+#define MAX_Q 100 // maximum baseQuality phred score
 
-Logger* Logger::gLogger = NULL;
-
-std::string bimChr2ncbiChr(std::string & bimChr) {
-  int nChr = atoi(bimChr.c_str());
-  std::string ncbiChr;
-  if ( nChr == 0 ) {
-    Logger::gLogger->error("Unrecognized chromosome %s",bimChr.c_str());
-  }
-  else if ( nChr < 23 ) {
-    ncbiChr = bimChr;
-  }
-  else if ( nChr == 23 ) {
-    ncbiChr = "X";
-  }
-  else if ( nChr == 24 ) {
-    ncbiChr = "Y";
-  }
-  else if ( nChr == 25 ) {
-    ncbiChr = "XY";
-  }
-  else if ( nChr == 26 ) {
-    ncbiChr = "MT";
-  }
-  else {
-    Logger::gLogger->error("Unrecognized chromosome %s",bimChr.c_str());
-  }
-  return ncbiChr;
-}
+Logger* Logger::gLogger = NULL; // Message log 
 
 int main(int argc, char** argv) {
   // configurable parameters
   std::string sRef, sInFile, sInPrefix, sInSuffix, sIndexFile, sOutFile, sLogFile, sBfile, sBimFile;
-  bool bVerbose, bNoEOF, bSelfOnly, bBimAF, bMemoryMap;
+  bool bVerbose, bNoEOF, bSelfOnly, bBimAF, bMemoryMap, bUCSC, bPrecise;
   uint32_t minQ, maxQ, minMapQ, maxDepth;
   double genoError, minAF, mixUnit, ibdUnit;
 
-  // parameters not set to be configuraable
+  // parameters currently not set to be configuraable
   bool sameSMFlag = true;
   uint16_t includeSamFlag = 0x0000;
   uint16_t excludeSamFlag = 0x0704;
   
+  // configure parameter settings
   try {
-    TCLAP::CmdLine cmd("Command description message", ' ', "0.01");
+    TCLAP::CmdLine cmd("Command description message", ' ', "0.0.2");
     TCLAP::ValueArg<std::string> argReference("r","reference","Karma's reference sequence",true,"","string");
     TCLAP::ValueArg<std::string> argIn("i","in","Input BAM file. Must be sorted and indexed",false,"","string");
     TCLAP::ValueArg<std::string> argInprefix("p","inprefix","Prefix of input BAM file for multi-chromosome inputs",false,"","string");
@@ -103,6 +75,8 @@ int main(int argc, char** argv) {
     TCLAP::SwitchArg switchSelfOnly("S","selfonly","compare the genotypes with SELF (annotated sample) only (default:off)",cmd,false);
     TCLAP::SwitchArg switchBimAF("F","bimAF","use the allele frequency information by loading .bimp file instead of .bim file",cmd,false);
     TCLAP::SwitchArg switchMemoryMap("M","mmap","toggle whether to use memory map (default:true)",cmd,true);
+    TCLAP::SwitchArg switchUCSC("u","ucsc","use UCSC-type reference file, where the sequence name starts with a prefix 'chr'",cmd,false);
+    TCLAP::SwitchArg switchPrecise("","precise","improve the precision of calculation for high-depth (>100x) data",cmd,false);
     
     cmd.add(argReference);
     cmd.add(argIn);
@@ -146,43 +120,54 @@ int main(int argc, char** argv) {
     bSelfOnly = switchSelfOnly.getValue();
     bBimAF = switchBimAF.getValue();
     bMemoryMap = switchMemoryMap.getValue();
+    bUCSC = switchUCSC.getValue();
+    bPrecise = switchPrecise.getValue();
 
+    // create log file
     if ( sLogFile.empty() ) {
       sLogFile = sOutFile + ".log";
     }
     Logger::gLogger = new Logger(sLogFile.c_str(), bVerbose);
 
+    // print out effective arguments
     Logger::gLogger->writeLog("Arguments in effect: \n%s",cmd.toString().c_str());
   }
+  // abort if failed paring the arguments
   catch (TCLAP::ArgException &e)  { 
     std::cerr << "error: " << e.error() << " for arg " << e.argId() << std::endl; 
     abort();
   }
 
+  // post-process arguments to get configurable parameters
   uint32_t numIBDs = static_cast<uint32_t>(ceil(1./ibdUnit))+1;
   uint32_t numMixFracs = static_cast<uint32_t>(ceil(.5/mixUnit))+1;
   double* fIBDs = new double[numIBDs];
   double* fMixFracs = new double[numMixFracs];
   double fPhred2Err[MAX_Q+1];
 
+  // initialize the list of IBD values to compare
   for(uint32_t i=0; i < numIBDs; ++i) {
     fIBDs[i] = ibdUnit * static_cast<double>(i);
     if ( fIBDs[i] > 1. ) fIBDs[i] = 1.;
   }
+  // initialize the list of %MIX values to compare
   for(uint32_t i=0; i < numMixFracs; ++i) {
     fMixFracs[i] = mixUnit * static_cast<double>(i);
     if ( fMixFracs[i] > .5 ) fMixFracs[i] = 0.5;
   }
-
+  // precompute the conversion from Phred score to baseError
   for(int32_t i=0; i < MAX_Q+1; ++i) {
+    // If Phred>=maxQ, assume that the base quality is
+    // overestimated and apply an upper threshold.
     if ( i > static_cast<int32_t>(maxQ) ) {
-      fPhred2Err[i] = fPhred2Err[maxQ]; // Phred>=40 is uninformative
+      fPhred2Err[i] = fPhred2Err[maxQ]; 
     }
     else {
       fPhred2Err[i] = pow(10.,(0-i)/10.);
     }
   }
 
+  // check the validity of input files
   if ( sBfile.empty() && sBimFile.empty() ) {
     Logger::gLogger->error("Either --bfile or --bimfile is required");
   }
@@ -195,14 +180,15 @@ int main(int argc, char** argv) {
     Logger::gLogger->error("--in and (--inprefix & --insuffix) cannot be combined together");
   }
 
-  std::map<std::string,uint32_t> msRGidx;
-  std::vector<std::string> vsRGIDs;
-  std::vector<std::string> vsSMIDs;
-  std::string sBamSMID;
+  std::map<std::string,uint32_t> msRGidx; // map from RGID to an index
+  std::vector<std::string> vsRGIDs; // list of indexed RGID
+  std::vector<std::string> vsSMIDs; // list of SMID
+  std::string sBamSMID; // per-BAM sample ID, when all readGroups have consistent sample ID
 
-  GenomeSequence genomeSequence;
+  GenomeSequence genomeSequence; // genomeSequence object
   genomeSequence.setReferenceName(sRef.c_str());
   genomeSequence.useMemoryMap(bMemoryMap);
+  // open and create one if the index does not exist
   if ( genomeSequence.open() ) {
     Logger::gLogger->warning("Failed to open karma-type index of reference file %s. Creating a new one..",sRef.c_str());
     if ( genomeSequence.create(false)) {
@@ -213,16 +199,21 @@ int main(int argc, char** argv) {
     }
   }
 
+  // open genotype file
   BedMatrix* pBedMatrix = new BedMatrix;
   if ( !sBfile.empty() ) {
-    if ( bBimAF ) {
+    if ( bBimAF ) { 
+      // genotype is available, but use estimated allele frequency
+      // this is useful when genotype data has limited # of individuals
+      // (e.g. with --selfOnly option)
       pBedMatrix->open((sBfile+".fam").c_str(),(sBfile+".bimp").c_str(),(sBfile+".bed").c_str());
     }
     else {
       pBedMatrix->open((sBfile+".fam").c_str(),(sBfile+".bim").c_str(),(sBfile+".bed").c_str());
     }
   }
-  else if ( !sBimFile.empty() ) {
+  else if ( !sBimFile.empty() ) { 
+    // in the case only allele frequency is available
     bBimAF = true;
     pBedMatrix->openMarkerOnly(sBimFile.c_str());
   }
@@ -255,18 +246,19 @@ int main(int argc, char** argv) {
     BgzfFileType::setRequireEofBlock(false);
   }
 
-  std::string curChrom;
-  if ( !sInPrefix.empty() ) {
+  std::string curChrom; // variable needed for tracking multi-chromosome data
+  if ( !sInPrefix.empty() ) { // set up the first file to open
     if ( !sIndexFile.empty() ) {
       Logger::gLogger->error("Cannot use --index option with --prefix and --suffix option");
     }
 
     // open current BAM file
     curChrom = pBedMatrix->vpGenMarkers[0]->sChrom;
-    sInFile = sInPrefix + bimChr2ncbiChr(curChrom) + sInSuffix;
-    sIndexFile = sInFile + ".bai";
+    sInFile = sInPrefix + pBedMatrix->vpGenMarkers[0]->getNCBIChromosomeName() + sInSuffix;
+    sIndexFile = sInFile + ".bai"; // index name is fixed with multi-chromosome case
   }
 
+  // routines for reading headers of the BAM file
   SamFile inBam;
   SamFileHeader inHeader;
   if ( ! (inBam.OpenForRead(sInFile.c_str()))  ) {
@@ -274,17 +266,22 @@ int main(int argc, char** argv) {
   }
   inBam.ReadHeader(inHeader);
 
+  // set index name
   if ( sIndexFile.empty() ) {
     sIndexFile = sInFile + ".bai";
   }
 
+  // read index file
   if ( ! inBam.ReadBamIndex( sIndexFile.c_str() ) ) {
     Logger::gLogger->error("Cannot open BAM file index %s for reading - %s", sIndexFile.c_str());
   }
 
   Logger::gLogger->writeLog("Reading header Records");
 
+  // read through header records to get the readGroup ID and corresponding sample ID. 
+  // It also checkes whether the sequenceName matches with the specified convention
   SamHeaderRecord* pSamHeaderRecord;
+  bool bObservedChrinSQ = false;
   while( (pSamHeaderRecord = inHeader.getNextHeaderRecord()) != NULL ) {
     if ( pSamHeaderRecord->getType() == SamHeaderRecord::RG ) {
       std::string sRGID = pSamHeaderRecord->getTagValue("ID");
@@ -308,12 +305,23 @@ int main(int argc, char** argv) {
       uint32_t idx = msRGidx.size();
       msRGidx[sRGID] = idx;
     }
+    else if ( pSamHeaderRecord->getType() == SamHeaderRecord::SQ ) {
+      std::string sSN = pSamHeaderRecord->getTagValue("SN");
+      if ( strncmp(sSN.c_str(), "chr", 3) == 0 ) {
+	bObservedChrinSQ = true;
+      }
+    }
   }
 
-  uint32_t nCompInds = 0;
-  int32_t nSelfIndex = -1;
+  if ( bObservedChrinSQ != bUCSC ) {
+    Logger::gLogger->warning("--ucsc option %s specificed but %s 'chr' in the sequence names. If the results contain no data, try to toggle the option", bUCSC ? "is" : "is not", bObservedChrinSQ ? "observed" : "not observed");
+  }
+
+  uint32_t nCompInds = 0; // number of individuals to compare
+  int32_t nSelfIndex = -1; // set when --selfonly option is set and the corresponding individual is found
   if ( bSelfOnly ) {
     if ( sameSMFlag ) {
+      // when --selfonly is set, copy the matching individual to the very beginning of the data
       for(uint32_t i=0; i < nInds; ++i) {
 	if ( sBamSMID.compare(pBedMatrix->vpIndividuals[i]->sIndID) == 0 ) {
 	  if ( nSelfIndex < 0 ) {
@@ -325,6 +333,7 @@ int main(int argc, char** argv) {
 	  }
 	}
       }
+      // disregard the genotypes when matching individual was not found with --selfonly option
       if ( nSelfIndex < 0 ) {
 	Logger::gLogger->warning("--selfOnly was used but the individual in the BAM file cannot be found in the genotype file, skupping genotype comparison");
 	nCompInds = 0;
@@ -346,8 +355,6 @@ int main(int argc, char** argv) {
     Logger::gLogger->writeLog("\t%d: %s",i+1,vsRGIDs[i].c_str());
   }
   
-  //SamStatus::Status returnStatus = SamStatus::SUCCESS;
-
   SamRecord samRecord;
   std::vector<uint32_t> nRGIndices;
   std::vector<char> cBases;
@@ -372,116 +379,119 @@ int main(int argc, char** argv) {
   // (#ind) * (#geno) matrix of non-reference count given marker genotype  
   uint32_t* nSMIndGenoMarkerAlts = new uint32_t[nCompInds * 3]();
 
-  // (#RG) * (#ind) * (#ind) * (#LLKs)
+  // genotype-sequence IBD likelioods per RG, and per SM
   double* fSumRGIndLLKs = new double[nRGs * nCompInds * numIBDs]();
-  double* fRGGenoIBDLKs = new double[nRGs * 4]();
-  double* fRGnonIBDLKs = new double[nRGs]();
+  double* fRGGenoIBDLLKs = new double[nRGs * 3]();
+  //double* fRGnonIBDLLKs = new double[nRGs]();
 
   double* fSumSMIndLLKs = new double[nCompInds * numIBDs]();
-  double* fSMGenoIBDLKs = new double[4]();
-  double fSMnonIBDLK = 1.;
+  double* fSMGenoIBDLLKs = new double[3]();
+  //double fSMnonIBDLLK = (bPrecise ? 0 : 1.);
 
+  // %MIX likelihood per RG, and per SM
   double* fRGMixLLKs = new double[nRGs * numMixFracs]();
   double* fSMMixLLKs = new double[numMixFracs]();
-  double* fRGMixMarkerLKs = new double[nRGs * 9];
-  double* fSMMixMarkerLKs = new double[9];
-  //double* fRGMixMarkerLKs = new double[nRGs * numMixFracs];
-  //double* fSMMixMarkerLKs = new double[numMixFracs];
+  double* fRGMixMarkerLLKs = new double[nRGs * 9];
+  double* fSMMixMarkerLLKs = new double[9];
 
+  // Depth distribution of the markers per RG
   uint32_t* nRGMarkerDepth = new uint32_t[nRGs * (maxDepth+1)]();
   uint32_t* nRGMarkerDepthRefOnly = new uint32_t[nRGs * (maxDepth+1)]();
   uint32_t* nRGMarkerDepthAltOnly = new uint32_t[nRGs * (maxDepth+1)]();
   uint32_t* nRGMarkerDepthRefAlt = new uint32_t[nRGs * (maxDepth+1)]();
   double* fRGSumEHetDepth = new double[nRGs * (maxDepth+1)]();
 
+  // count of ref, alt, other bases per readgroup
   uint32_t* nRGRefCnts = new uint32_t[nRGs]();
   uint32_t* nRGAltCnts = new uint32_t[nRGs]();
   uint32_t* nRGEtcCnts = new uint32_t[nRGs]();
 
+  // Depth distribution of the markers per SM
   uint32_t* nSMMarkerDepth = new uint32_t[(maxDepth+1)]();
   uint32_t* nSMMarkerDepthRefOnly = new uint32_t[(maxDepth+1)]();
   uint32_t* nSMMarkerDepthAltOnly = new uint32_t[(maxDepth+1)]();
   uint32_t* nSMMarkerDepthRefAlt = new uint32_t[(maxDepth+1)]();
   double* fSMSumEHetDepth = new double[(maxDepth+1)]();
 
+  // count of ref, alt, other bases per SM
   uint32_t nSMRefCnt = 0;
   uint32_t nSMAltCnt = 0;
   uint32_t nSMEtcCnt = 0;
 
   Logger::gLogger->writeLog("Iterating over the genotypes..");
-  //for( uint32_t markerCnt = 0; pBedMatrix->iterateMarker() && (markerCnt < 1000 ); ++markerCnt ) {
+  //for( uint32_t markerCnt = 0; pBedMatrix->iterateMarker() && (markerCnt < 1000 ); ++markerCnt ) { // left for debugging purpose
 
   // iterate each variable sites (marker)
   for( uint32_t markerCnt = 0; pBedMatrix->iterateMarker(); ++markerCnt ) {
     if ( (markerCnt+1) % 1000 == 0 )
       Logger::gLogger->writeLog("Processing %u / %u markers",markerCnt+1,nMarkers);
     
-    // In 1000G files, one BAM file is splitted into each chromosome, this is the routine
-    // to process those special cases
-    if ( ( !sInPrefix.empty() ) 
-	 && ( pBedMatrix->currentMarker()->sChrom.compare(curChrom) != 0 ) )  {
-      inBam.Close();
-      
+    // In 1000G data, one BAM file is splitted into each chromosome, 
+    // this is the routine to process the special cases
+    // [inPrefix][ncbiChrName][outPrefix]
+    if ( pBedMatrix->currentMarker()->sChrom.compare(curChrom) != 0 ) {
       curChrom = pBedMatrix->currentMarker()->sChrom;
-      std::string ncbiChr = bimChr2ncbiChr(pBedMatrix->currentMarker()->sChrom);
-      Logger::gLogger->writeLog("Opening chromosome %s",ncbiChr.c_str());
+      Logger::gLogger->writeLog("Opening chromosome %s",pBedMatrix->currentMarker()->getNCBIChromosomeName().c_str());
+      
+      if ( !sInPrefix.empty() )  {
+	// if the current marker does not match to current BAM file,
+	// close the current file, open new file with header
+	inBam.Close();
+	
+	sInFile = sInPrefix + pBedMatrix->currentMarker()->getNCBIChromosomeName() + sInSuffix;
+	if ( ! (inBam.OpenForRead(sInFile.c_str())) ) {
+	  Logger::gLogger->error("Cannot open BAM file %s for reading - %s",sInFile.c_str(), SamStatus::getStatusString(inBam.GetStatus()) );	
+	}
+	inBam.ReadHeader(inHeader);
+	inHeader.resetHeaderRecordIter();
+	sIndexFile = sInFile + ".bai";
 
-      sInFile = sInPrefix + ncbiChr + sInSuffix;
-      if ( ! (inBam.OpenForRead(sInFile.c_str())) ) {
-	Logger::gLogger->error("Cannot open BAM file %s for reading - %s",sInFile.c_str(), SamStatus::getStatusString(inBam.GetStatus()) );	
-      }
-      inBam.ReadHeader(inHeader);
-      inHeader.resetHeaderRecordIter();
-      sIndexFile = sInFile + ".bai";
-
-      if ( ! inBam.ReadBamIndex( sIndexFile.c_str() ) ) {
-	Logger::gLogger->error("Cannot open BAM file index %s for reading - %s", sIndexFile.c_str());	
-      }
-
-      //Logger::gLogger->writeLog("foo");
-
-      for( uint32_t k=0, rgCnt = 0; (pSamHeaderRecord = inHeader.getNextHeaderRecord()) != NULL; ++k ) {
-	if ( pSamHeaderRecord->getType() == SamHeaderRecord::RG ) {
-	  std::string sRGID = pSamHeaderRecord->getTagValue("ID");
-	  std::string sSMID = pSamHeaderRecord->getTagValue("SM");
-	  //Logger::gLogger->writeLog("bar %u %u %s %s",k,rgCnt,sRGID.c_str(),sSMID.c_str());
-
-	  if ( sRGID.empty() ) {
-	    Logger::gLogger->error("Readgroup ID is empty");
+	if ( ! inBam.ReadBamIndex( sIndexFile.c_str() ) ) {
+	  Logger::gLogger->error("Cannot open BAM file index %s for reading - %s", sIndexFile.c_str());	
+	}
+	
+	for( uint32_t k=0, rgCnt = 0; (pSamHeaderRecord = inHeader.getNextHeaderRecord()) != NULL; ++k ) {
+	  if ( pSamHeaderRecord->getType() == SamHeaderRecord::RG ) {
+	    std::string sRGID = pSamHeaderRecord->getTagValue("ID");
+	    std::string sSMID = pSamHeaderRecord->getTagValue("SM");
+	    
+	    if ( sRGID.empty() ) {
+	      Logger::gLogger->error("Readgroup ID is empty");
+	    }
+	    if ( sSMID.empty() ) {
+	      Logger::gLogger->warning("SM tag is missing in read group %s",sRGID.c_str());
+	    }
+	    
+	    if ( vsRGIDs[rgCnt].compare(sRGID) != 0 ) {
+	      Logger::gLogger->error("Nonidentical headers - ReadGroup ID is different at %u : %s vs %s",rgCnt,vsRGIDs[rgCnt].c_str(),sRGID.c_str());
+	    }
+	    if ( vsSMIDs[rgCnt].compare(sSMID) != 0 ) {
+	      Logger::gLogger->error("Nonidentical headers - Readgroup SM is different at %u : %s vs %s",rgCnt,vsSMIDs[rgCnt].c_str(),sSMID.c_str());
+	    }
+	    ++rgCnt;
 	  }
-	  if ( sSMID.empty() ) {
-	    Logger::gLogger->warning("SM tag is missing in read group %s",sRGID.c_str());
-	  }
-
-	  if ( vsRGIDs[rgCnt].compare(sRGID) != 0 ) {
-	    Logger::gLogger->error("Nonidentical headers - ReadGroup ID is different at %u : %s vs %s",rgCnt,vsRGIDs[rgCnt].c_str(),sRGID.c_str());
-	  }
-	  if ( vsSMIDs[rgCnt].compare(sSMID) != 0 ) {
-	    Logger::gLogger->error("Nonidentical headers - Readgroup SM is different at %u : %s vs %s",rgCnt,vsSMIDs[rgCnt].c_str(),sSMID.c_str());
-	  }
-	  ++rgCnt;
 	}
       }
-
-      //Logger::gLogger->writeLog("bar");
     }
 
 
-    // Routine for processing each marker
+    // Routine for processing genotype data at each marker
 
     // Read the genotype file
     char* bedGenos = pBedMatrix->currentGenotypes();
     char a1 = pBedMatrix->currentMarker()->sAllele1[0];
     char a2 = pBedMatrix->currentMarker()->sAllele2[0];
-    genomeIndex_t markerIndex = genomeSequence.getGenomePosition( pBedMatrix->currentMarker()->sChrom.c_str(), pBedMatrix->currentMarker()->nBasePosition );
+    genomeIndex_t markerIndex = genomeSequence.getGenomePosition( bUCSC ? pBedMatrix->currentMarker()->getUCSCChromosomeName().c_str() : pBedMatrix->currentMarker()->getNCBIChromosomeName().c_str(), pBedMatrix->currentMarker()->nBasePosition );
     char refBase = genomeSequence[markerIndex];
     bool flipFlag = false;
 
+    // routines for a1/a2 flip
+    // note that strand flip would not be automatically handled
     if ( refBase == a1 ) {
       // fine
     }
     else if ( refBase == a2 ) {
-      // flip genotypes
+      // flip genotypes if reference base does not match
       char tmp = a2;
       a2 = a1;
       a1 = tmp;
@@ -492,11 +502,9 @@ int main(int argc, char** argv) {
 	  bedGenos[i] = 1;
       }
       flipFlag = true;
-      //Logger::gLogger->writeLog("Marker alleles %s at %s:%u was flipped",pBedMatrix->currentMarker()->sMarkerID.c_str(), pBedMatrix->currentMarker()->sChrom.c_str(), pBedMatrix->currentMarker()->nBasePosition);
-
     }
     else {
-      Logger::gLogger->error("Marker alleles %s at %s:%u do not match to reference allele",pBedMatrix->currentMarker()->sMarkerID.c_str(), pBedMatrix->currentMarker()->sChrom.c_str(), pBedMatrix->currentMarker()->nBasePosition);
+      Logger::gLogger->error("Marker alleles %s at %s:%u do not match to reference allele",pBedMatrix->currentMarker()->sMarkerID.c_str(), pBedMatrix->currentMarker()->getNCBIChromosomeName().c_str(), pBedMatrix->currentMarker()->nBasePosition);
     }
 
     // calculate non-reference allele frequency
@@ -530,18 +538,20 @@ int main(int argc, char** argv) {
     //Logger::gLogger->writeLog("%c %c %c %lf",refBase,a1,a2,alleleFreq);
 
     // genotype-match likelihood
-    for(uint32_t i=0; i < nRGs*4; ++i) {
-      fRGGenoIBDLKs[i] = 1.;
-    }
-    for(uint32_t i=0; i < nRGs; ++i) {
-      fRGnonIBDLKs[i] = 1.;
+    for(uint32_t i=0; i < nRGs*3; ++i) {
+      fRGGenoIBDLLKs[i] = (bPrecise ? 0. : 1.);
     }
+    //for(uint32_t i=0; i < nRGs; ++i) {
+    //  fRGnonIBDLLKs[i] = (bPrecise ? 0. : 1.);
+    //}
 
-    fSMGenoIBDLKs[0] = fSMGenoIBDLKs[1] = fSMGenoIBDLKs[2] = fSMGenoIBDLKs[3] = fSMnonIBDLK = 1.; 
+    fSMGenoIBDLLKs[0] = fSMGenoIBDLLKs[1] = fSMGenoIBDLLKs[2] = (bPrecise ? 0. : 1.);
 
 
     // retrieve the bases aligned to the marker position
-    int refID = inHeader.GetReferenceID(pBedMatrix->currentMarker()->sChrom.c_str());
+    //std::cerr << "foo" << std::endl;
+    int refID = inHeader.GetReferenceID( bUCSC ? pBedMatrix->currentMarker()->getUCSCChromosomeName().c_str() : pBedMatrix->currentMarker()->getNCBIChromosomeName().c_str() );
+    //std::cerr << "bar" << std::endl;
     uint32_t pos = pBedMatrix->currentMarker()->nBasePosition; // 1-based position
     int numSectionRecords = 0;
     
@@ -647,7 +657,8 @@ int main(int argc, char** argv) {
       }*/
 
       if ( nRGIndices.size() <= maxDepth ) {
-	double probReadIBD[4], probReadnonIBD;
+	//double probReadIBD[4], probReadnonIBD;
+	double probReadIBD[3]; // missing genotype is not allowed
 	for(uint32_t i=0; i < nRGs; ++i) {
 	  nRGRefCnts[i] = 0;
 	  nRGAltCnts[i] = 0;
@@ -661,33 +672,46 @@ int main(int argc, char** argv) {
 	  double baseError = fPhred2Err[cQuals[i]-33];
 
 	  if ( cBases[i] == a1 ) {
-	    probReadIBD[0] = probReadnonIBD = (1.-alleleFreq)*(1.-baseError) + alleleFreq * (baseError/3.);
-	    probReadIBD[1] = 1.-baseError;
-	    probReadIBD[2] = 0.5-baseError/3.;
-	    probReadIBD[3] = baseError/3.;
+	    //probReadIBD[0] = probReadnonIBD = (1.-alleleFreq)*(1.-baseError) + alleleFreq * (baseError/3.);
+	    probReadIBD[0] = 1.-baseError; 
+	    probReadIBD[1] = 0.5-baseError/3.;
+	    probReadIBD[2] = baseError/3.;
 	    ++nRGRefCnts[nRGIndices[i]];
 	    ++nSMRefCnt;
 	  }
 	  else if ( cBases[i] == a2 ) {
-	    probReadIBD[0] = probReadnonIBD = (alleleFreq)*(1.-baseError) + (1.-alleleFreq) * (baseError/3.);
-	    probReadIBD[1] = baseError/3.;
-	    probReadIBD[2] = 0.5-baseError/3.;
-	    probReadIBD[3] = 1.-baseError;
+	    //probReadIBD[0] = probReadnonIBD = (alleleFreq)*(1.-baseError) + (1.-alleleFreq) * (baseError/3.);
+	    probReadIBD[0] = baseError/3.;
+	    probReadIBD[1] = 0.5-baseError/3.;
+	    probReadIBD[2] = 1.-baseError;
 	    ++nRGAltCnts[nRGIndices[i]];
 	    ++nSMAltCnt;
 	  }
 	  else {
-	    probReadIBD[0] = probReadIBD[1] = probReadIBD[2] = probReadIBD[3] = probReadnonIBD = baseError/3.;
+	    //probReadIBD[0] = probReadnonIBD
+	    probReadIBD[0] = probReadIBD[1] = probReadIBD[2] = baseError/3.;
 	    ++nRGEtcCnts[nRGIndices[i]];
 	    ++nSMEtcCnt;
 	  }
 
-	  for(uint32_t j=0; j < 4; ++j) {
-	    fRGGenoIBDLKs[nRGIndices[i]*4+j] *= probReadIBD[j];
-	    fSMGenoIBDLKs[j] *= probReadIBD[j];
+	  for(uint32_t j=0; j < 3; ++j) {
+	    if ( bPrecise ) {
+	      fRGGenoIBDLLKs[nRGIndices[i]*3+j] += log(probReadIBD[j]);
+	      fSMGenoIBDLLKs[j] += log(probReadIBD[j]);
+	    }
+	    else {
+	      fRGGenoIBDLLKs[nRGIndices[i]*3+j] *= probReadIBD[j];
+	      fSMGenoIBDLLKs[j] *= probReadIBD[j];
+	    }
 	  }
-	  fRGnonIBDLKs[nRGIndices[i]] *= probReadnonIBD;
-	  fSMnonIBDLK *= probReadnonIBD;
+	  //if ( bPrecise ) {
+	  //  fRGnonIBDLLKs[nRGIndices[i]] += log(probReadnonIBD);
+	  //  fSMnonIBDLLK += log(probReadnonIBD);
+	  //}
+	  //else {
+	  //  fRGnonIBDLLKs[nRGIndices[i]] *= probReadnonIBD;
+	  //  fSMnonIBDLLK *= probReadnonIBD;
+	  //}
 
 	  for(uint32_t j=0; j < nCompInds; ++j) {
 	    if ( bedGenos[j] > 0 ) {
@@ -748,22 +772,68 @@ int main(int argc, char** argv) {
 
 	// calculating per-RG LLK
 	for(uint32_t i=0; i < nRGs; ++i) {
-	  for(uint32_t j=0; j < nCompInds; ++j) {
-	    double probReadsIBD = (bedGenos[j] > 0) ? ( (1.-1.5*genoError) * fRGGenoIBDLKs[ i*4 + bedGenos[j] ] + 0.5*genoError*(fRGGenoIBDLKs[i*4+1]+fRGGenoIBDLKs[i*4+2]+fRGGenoIBDLKs[i*4+3]) ) : fRGGenoIBDLKs[i*4+0];
-	    for(uint32_t k=0; k < numIBDs; ++k) {
-	      double llk = log( fIBDs[k]*probReadsIBD + (1.-fIBDs[k])*fRGnonIBDLKs[i] );
+	  double fRGNonIBDLLK; // likelihood of reads given non-IBDs 
+	  if ( bPrecise ) {
+	    uint32_t maxIdx = 0;
+	    for(uint32_t j=1; j < 3; ++j) {
+	      if ( fRGGenoIBDLLKs[i*3+j] > fRGGenoIBDLLKs[i*3+maxIdx] ) {
+		maxIdx = j;
+	      }
+	    }
+	    fRGNonIBDLLK = fRGGenoIBDLLKs[i*3+maxIdx] + log((1.-alleleFreq)*(1.-alleleFreq)*exp(fRGGenoIBDLLKs[i*3+0]-fRGGenoIBDLLKs[i*3+maxIdx]) + 2.*alleleFreq*(1.-alleleFreq)*exp(fRGGenoIBDLLKs[i*3+1]-fRGGenoIBDLLKs[i*3+maxIdx]) + alleleFreq*alleleFreq*exp(fRGGenoIBDLLKs[i*3+2]-fRGGenoIBDLLKs[i*3+maxIdx]));
+	  }
+	  else {
+	    fRGNonIBDLLK = (1.-alleleFreq)*(1.-alleleFreq)*fRGGenoIBDLLKs[i*3+0] + 2.*alleleFreq*(1.-alleleFreq)*fRGGenoIBDLLKs[i*3+1] + alleleFreq*alleleFreq*fRGGenoIBDLLKs[i*3+2];
+	  }
 
-	      fSumRGIndLLKs[i*nCompInds*numIBDs+j*numIBDs+k] += llk;
+	  for(uint32_t j=0; j < nCompInds; ++j) {
+	    if ( bPrecise ) {
+	      double probReadsIBD = (bedGenos[j] > 0) ? ( (1.-1.5*genoError) * exp(fRGGenoIBDLLKs[ i*3 + bedGenos[j] - 1] - fRGNonIBDLLK)  + 0.5*genoError*(exp(fRGGenoIBDLLKs[i*3+0] - fRGNonIBDLLK) + exp(fRGGenoIBDLLKs[i*3+1] - fRGNonIBDLLK) + exp(fRGGenoIBDLLKs[i*3+2] - fRGNonIBDLLK) ) ) : 1.;
+	      for(uint32_t k=0; k < numIBDs; ++k) {
+		double llk = fRGNonIBDLLK + log( fIBDs[k]*probReadsIBD + (1.-fIBDs[k])*1. );
+		
+		fSumRGIndLLKs[i*nCompInds*numIBDs+j*numIBDs+k] += llk;
+	      }
+	    }
+	    else {
+	      double probReadsIBD = (bedGenos[j] > 0) ? ( (1.-1.5*genoError) * ( fRGGenoIBDLLKs[ i*3 + bedGenos[j] - 1 ] )  + 0.5*genoError*(fRGGenoIBDLLKs[i*3+0] + fRGGenoIBDLLKs[i*3+1] + fRGGenoIBDLLKs[i*3+2]) ) : fRGNonIBDLLK;
+	      for(uint32_t k=0; k < numIBDs; ++k) {
+		double llk = log( fIBDs[k]*probReadsIBD + (1.-fIBDs[k])*fRGNonIBDLLK );
+		fSumRGIndLLKs[i*nCompInds*numIBDs+j*numIBDs+k] += llk;
+	      }
 	    }
 	  }
 	}
 
 	// calcuating per-SM LLK
+	double fSMNonIBDLLK;
+	if ( bPrecise ) {
+	  uint32_t maxIdx = 0;
+	  for(uint32_t j=1; j < 3; ++j) {
+	    if ( fSMGenoIBDLLKs[j] > fSMGenoIBDLLKs[maxIdx] ) {
+	      maxIdx = j;
+	    }
+	  }
+	  fSMNonIBDLLK = fSMGenoIBDLLKs[maxIdx] + log((1.-alleleFreq)*(1.-alleleFreq)*exp(fSMGenoIBDLLKs[0]-fSMGenoIBDLLKs[maxIdx]) + 2.*alleleFreq*(1.-alleleFreq)*exp(fSMGenoIBDLLKs[1]-fSMGenoIBDLLKs[maxIdx]) + alleleFreq*alleleFreq*exp(fSMGenoIBDLLKs[2]-fSMGenoIBDLLKs[maxIdx]));
+	}
+	else {
+	  fSMNonIBDLLK = (1.-alleleFreq)*(1.-alleleFreq)*fSMGenoIBDLLKs[0] + 2.*alleleFreq*(1.-alleleFreq)*fSMGenoIBDLLKs[1] + alleleFreq*alleleFreq*fSMGenoIBDLLKs[2];
+	}
+
 	for(uint32_t j=0; j < nCompInds; ++j) {
-	  double probReadsIBD = (bedGenos[j] > 0) ? ( (1.-1.5*genoError) * fSMGenoIBDLKs[ bedGenos[j] ] + 0.5*genoError*(fSMGenoIBDLKs[1]+fSMGenoIBDLKs[2]+fSMGenoIBDLKs[3]) ) : fSMGenoIBDLKs[0];
-	  for(uint32_t k=0; k < numIBDs; ++k) {
-	    double llk = log( fIBDs[k]*probReadsIBD + (1.-fIBDs[k])*fSMnonIBDLK );
-	    fSumSMIndLLKs[j*numIBDs+k] += llk;
+	  if ( bPrecise ) {
+	    double probReadsIBD = (bedGenos[j] > 0) ? ( (1.-1.5*genoError) * exp(fSMGenoIBDLLKs[ bedGenos[j] - 1] - fSMNonIBDLLK)  + 0.5*genoError*(exp(fSMGenoIBDLLKs[0] - fSMNonIBDLLK) + exp(fSMGenoIBDLLKs[1] - fSMNonIBDLLK) + exp(fSMGenoIBDLLKs[2] - fSMNonIBDLLK) ) ) : 1.;
+	    for(uint32_t k=0; k < numIBDs; ++k) {
+	      double llk = fSMNonIBDLLK + log( fIBDs[k]*probReadsIBD + (1.-fIBDs[k])*1. );
+	      fSumSMIndLLKs[j*numIBDs+k] += llk;
+	    }
+	  }
+	  else {
+	    double probReadsIBD = (bedGenos[j] > 0) ? ( (1.-1.5*genoError) * ( fSMGenoIBDLLKs[ bedGenos[j] - 1 ] )  + 0.5*genoError*(fSMGenoIBDLLKs[0] + fSMGenoIBDLLKs[1] + fSMGenoIBDLLKs[2]) ) : fSMNonIBDLLK;
+	    for(uint32_t k=0; k < numIBDs; ++k) {
+	      double llk = log( fIBDs[k]*probReadsIBD + (1.-fIBDs[k])*fSMNonIBDLLK );
+	      fSumSMIndLLKs[j*numIBDs+k] += llk;
+	    }
 	  }
 	}
 
@@ -781,10 +851,10 @@ int main(int argc, char** argv) {
 	for(uint32_t j=0; j < numMixFracs; ++j) {
 	  double alpha = fMixFracs[j];
 	  for(uint32_t i=0; i < 9; ++i) {
-	    fSMMixMarkerLKs[i] = 1.;
+	    fSMMixMarkerLLKs[i] = (bPrecise ? 0. : 1.);
 	  }
 	  for(uint32_t i=0; i < 9*nRGs; ++i) {
-	    fRGMixMarkerLKs[i] = 1.;
+	    fRGMixMarkerLLKs[i] = (bPrecise ? 0. : 1.);
 	  }
 	  //double valuesToMultiply[9] = {1.,1.,1.,1.,1.,1.,1.,1.,1.};
 	  double lks[9];
@@ -826,31 +896,75 @@ int main(int argc, char** argv) {
 	    }
 
 	    for(uint32_t k=0; k < 9; ++k) {
-	      //valuesToMultiply[k] *= lks[k];
-	      fSMMixMarkerLKs[k] *= lks[k];
-	      fRGMixMarkerLKs[nRGIndices[i] * 9 + k] *= lks[k];
+	      if ( bPrecise ) {
+		//valuesToMultiply[k] *= lks[k];
+		fSMMixMarkerLLKs[k] += log(lks[k]);
+		fRGMixMarkerLLKs[nRGIndices[i] * 9 + k] += log(lks[k]);
+	      }
+	      else {
+		fSMMixMarkerLLKs[k] *= lks[k];
+		fRGMixMarkerLLKs[nRGIndices[i] * 9 + k] *= lks[k];
+	      }
 	    }
 	  }
-	  
+
 	  double perMarkerProb = 0;
-	  for(uint32_t k1 = 0; k1 < 3; ++k1) {
-	    for(uint32_t k2 = 0; k2 < 3; ++k2) {
-	      perMarkerProb += (fSMMixMarkerLKs[k1*3+k2]*genoFreq[k1]*genoFreq[k2]);
+	  if ( bPrecise ) {
+	    uint32_t maxIdx = 0;
+	    for(uint32_t k = 1; k < 9; ++k) {
+	      if ( fSMMixMarkerLLKs[maxIdx] < fSMMixMarkerLLKs[k] ) {
+		maxIdx = k;
+	      }
 	    }
+	    for(uint32_t k1 = 0; k1 < 3; ++k1) {
+	      for(uint32_t k2 = 0; k2 < 3; ++k2) {
+		perMarkerProb += (exp(fSMMixMarkerLLKs[k1*3+k2]-fSMMixMarkerLLKs[maxIdx])*genoFreq[k1]*genoFreq[k2]);
+	      }
+	    }
+	    fSMMixLLKs[j] += (fSMMixMarkerLLKs[maxIdx]+log(perMarkerProb));
+	  }
+	  else {
+	    for(uint32_t k1 = 0; k1 < 3; ++k1) {
+	      for(uint32_t k2 = 0; k2 < 3; ++k2) {
+		perMarkerProb += (fSMMixMarkerLLKs[k1*3+k2]*genoFreq[k1]*genoFreq[k2]);
+	      }
+	    }
+	    fSMMixLLKs[j] += log(perMarkerProb);
 	  }
-	  fSMMixLLKs[j] += log(perMarkerProb);
 
 	  for(uint32_t k=0; k < nRGs; ++k) {
 	    perMarkerProb = 0;
-	    for(uint32_t k1 = 0; k1 < 3; ++k1) {
-	      for(uint32_t k2 = 0; k2 < 3; ++k2) {
-		perMarkerProb += (fRGMixMarkerLKs[k*9+k1*3+k2]*genoFreq[k1]*genoFreq[k2]);
+	    if ( bPrecise ) {
+	      uint32_t maxIdx = 0;
+	      for(uint32_t k1 = 1; k1 < 9; ++k1) {
+		if ( fRGMixMarkerLLKs[k*9+maxIdx] < fRGMixMarkerLLKs[k*9+k1] ) {
+		  maxIdx = k1;
+		}
 	      }
+
+	      for(uint32_t k1 = 0; k1 < 3; ++k1) {
+		for(uint32_t k2 = 0; k2 < 3; ++k2) {
+		  //perMarkerProb += (fRGMixMarkerLKs[k*9+k1*3+k2]*genoFreq[k1]*genoFreq[k2]);
+		  perMarkerProb += (exp(fRGMixMarkerLLKs[k*9+k1*3+k2]-fRGMixMarkerLLKs[k*9+maxIdx])*genoFreq[k1]*genoFreq[k2]);
+		}
+	      }
+	      //fRGMixLLKs[k*numMixFracs + j] += log(perMarkerProb);
+	      fRGMixLLKs[k*numMixFracs + j] += (fRGMixMarkerLLKs[k*9+maxIdx]+log(perMarkerProb));
+	    }
+	    else {
+	      for(uint32_t k1 = 0; k1 < 3; ++k1) {
+		for(uint32_t k2 = 0; k2 < 3; ++k2) {
+		  perMarkerProb += (fRGMixMarkerLLKs[k*9+k1*3+k2]*genoFreq[k1]*genoFreq[k2]);
+		}
+	      }
+	      fRGMixLLKs[k*numMixFracs + j] += log(perMarkerProb);
 	    }
-	    fRGMixLLKs[k*numMixFracs + j] += log(perMarkerProb);
 	  }
 	}
       }
+      else {
+	Logger::gLogger->writeLog("Skipping marker %s at %s:%d due to excessive depth %d",pBedMatrix->currentMarker()->sMarkerID.c_str(), bUCSC ? pBedMatrix->currentMarker()->getUCSCChromosomeName().c_str() : pBedMatrix->currentMarker()->getNCBIChromosomeName().c_str(), pBedMatrix->currentMarker()->nBasePosition, nRGIndices.size());
+      }
     }
     else {
       //abort();
